# coding: utf-8
import json

import pytest
import responses

from datarobot import PredictionDataset, Project
from datarobot.utils import datetime_to_string, parse_time
from tests.test_helpers import fixture_file_path
from tests.utils import assert_equal_py2, assert_equal_py3, request_body_to_json


@pytest.fixture
def list_dataset_json():
    return """
    {"count": 1,
     "next": null,
     "previous": null,
     "data": [
         {
             "id": "ds-id",
             "projectId": "p-id",
             "name": "My PredictionDataset",
             "created": "2016-02-16T12:00:00.123456Z",
             "numRows": 100,
             "numColumns": 20,
             "forecastPoint": null,
             "relaxKnownInAdvanceFeaturesCheck": true,
             "dataQualityWarnings": {"has_kia_missing_values_in_forecast_window": false,
                                     "insufficient_rows_for_evaluating_models": false,
                                     "single_class_actual_value_column": false
                                     },
             "actualValueColumn": null,
             "detectedActualValueColumns": null,
             "containsTargetValues": true
         }
     ]
    }
    """


@pytest.fixture
def dataset_json(list_dataset_json):
    data = json.loads(list_dataset_json)
    return json.dumps(data["data"][0])


@pytest.fixture
def dataset_with_relax_kia_check_server_data(dataset_json):
    dataset = dict(json.loads(dataset_json))
    dataset["relaxKnownInAdvanceFeaturesCheck"] = True
    return dataset


@pytest.fixture
def dataset_with_forecast_point_server_data(dataset_json):
    dataset = dict(json.loads(dataset_json))
    dataset["forecastPoint"] = "2017-01-01T15:00:00Z"
    return dataset


@pytest.fixture
def dataset_with_millis_forecast_point_server_data(dataset_json):
    dataset = dict(json.loads(dataset_json))
    dataset["forecastPoint"] = "2017-01-01T15:00:00.800000Z"
    return dataset


@pytest.fixture
def dataset_with_start_end_server_data(dataset_json):
    dataset = dict(json.loads(dataset_json))
    dataset["predictionsStartDate"] = "2017-01-01T15:00:00Z"
    dataset["predictionsEndDate"] = "2017-06-01T15:00:00Z"
    dataset["actualValueColumn"] = "actuals"
    dataset["detectedActualValueColumns"] = [{"missing_count": 25, "name": "actuals"}]
    return dataset


@pytest.fixture
def dataset_with_millis_start_end_server_data(dataset_json):
    dataset = dict(json.loads(dataset_json))
    dataset["predictionsStartDate"] = "2017-01-01T15:00:00.800000Z"
    dataset["predictionsEndDate"] = "2017-06-01T15:00:00.800000Z"
    return dataset


@pytest.fixture
def dataset_uploaded_without_forecast_point_server_data(dataset_json):
    dataset = dict(json.loads(dataset_json))
    dataset["forecastPoint"] = "2017-01-01T15:00:00Z"
    dataset["forecastPointRange"] = ["2017-01-01 14:00:00+00:00", "2017-01-01 16:00:00+00:00"]
    dataset["dataStartDate"] = "2017-01-01 10:00:00+00:00"
    dataset["dataEndDate"] = "2017-01-01 20:00:00+00:00"
    dataset["maxForecastDate"] = "2017-01-01 20:00:00+00:00"
    dataset["detectedActualValueColumns"] = [{"missing_count": 25, "name": "actuals"}]
    return dataset


@pytest.fixture
def dataset_with_config_id(dataset_json):
    dataset = dict(json.loads(dataset_json))
    dataset["secondaryDatasetsConfigId"] = "config-id"
    return dataset


dataset_response = {
    "id": "ds-id",
    "projectId": "p-id",
    "name": "My PredictionDataset",
    "created": "2016-02-16T12:00:00.123456Z",
    "numRows": 100,
    "numColumns": 20,
    "forecastPoint": None,
    "relax_known_in_advance_features_check": True,
    "dataQualityWarnings": {
        "has_kia_missing_values_in_forecast_window": False,
        "single_class_actual_value_column": False,
        "insufficient_rows_for_evaluating_models": False,
    },
    "containsTargetValues": True,
}

datasets_list_response = {"count": 1, "next": None, "previous": None, "data": [dataset_response]}


def prep_successful_upload_response(upload_type, body):
    responses.add(
        responses.POST,
        "https://host_name.com/projects/p-id/predictionDatasets/{}/".format(upload_type),
        body="",
        status=202,
        content_type="application/json",
        adding_headers={"Location": "https://host_name.com/status/status-id/"},
    )
    responses.add(
        responses.GET,
        "https://host_name.com/status/status-id/",
        status=303,
        body="",
        content_type="application/json",
        adding_headers={
            "Location": "https://host_name.com/projects/p-id/predictionDatasets/ds-id/"
        },
    )
    responses.add(
        responses.GET,
        "https://host_name.com/projects/p-id/predictionDatasets/ds-id/",
        status=200,
        body=body,
        content_type="application/json",
    )


def assert_matches_expected_dataset(actual_dataset):
    assert actual_dataset.project_id == dataset_response["projectId"]
    assert actual_dataset.id == dataset_response["id"]
    assert actual_dataset.name == dataset_response["name"]
    assert actual_dataset.created == dataset_response["created"]
    assert actual_dataset.num_rows == dataset_response["numRows"]
    assert actual_dataset.num_columns == dataset_response["numColumns"]
    assert (
        actual_dataset.relax_known_in_advance_features_check
        == dataset_response["relax_known_in_advance_features_check"]
    )
    assert actual_dataset.forecast_point is None
    assert actual_dataset.data_quality_warnings == dataset_response["dataQualityWarnings"]
    assert actual_dataset.contains_target_values == dataset_response["containsTargetValues"]


def test_future_proof():
    PredictionDataset.from_server_data(dict(dataset_response, new_key="future"))


@responses.activate
def test_list_datasets(list_dataset_json):
    responses.add(
        responses.GET,
        "https://host_name.com/projects/p-id/predictionDatasets/",
        body=list_dataset_json,
    )
    project = Project("p-id")
    datasets = project.get_datasets()
    assert len(datasets) == 1
    dataset = datasets[0]
    assert_matches_expected_dataset(dataset)


@responses.activate
def test_retrieve_dataset():
    responses.add(
        responses.GET,
        "https://host_name.com/projects/p-id/predictionDatasets/ds-id/",
        body=json.dumps(dataset_response),
    )
    dataset = PredictionDataset.get("p-id", "ds-id")
    assert_matches_expected_dataset(dataset)


@responses.activate
def test_dataset_repr():
    responses.add(
        responses.GET,
        "https://host_name.com/projects/p-id/predictionDatasets/ds-id/",
        body=json.dumps(dataset_response),
    )
    dataset = PredictionDataset.get("p-id", "ds-id")
    assert_equal_py2(repr(dataset), "PredictionDataset(u'My PredictionDataset')")
    assert_equal_py3(repr(dataset), "PredictionDataset('My PredictionDataset')")


def test_dataset_repr_unicode():
    dataset = PredictionDataset(
        "projectId",
        "datasetId",
        name=u"赤木",
        created="2016-02-16T12:00:00.123456Z",
        num_rows=100,
        num_columns=20,
    )
    assert_equal_py2(repr(dataset), "PredictionDataset(u'\\u8d64\\u6728')")
    assert_equal_py3(repr(dataset), "PredictionDataset('赤木')")


@responses.activate
def test_upload_dataset_from_url(dataset_json):
    prep_successful_upload_response("urlUploads", dataset_json)
    project = Project("p-id")
    file_url = "https://my_files.com/my_dataset.csv"
    new_dataset = project.upload_dataset(file_url)
    assert_matches_expected_dataset(new_dataset)
    request_body = request_body_to_json(responses.calls[0].request)
    assert request_body == {"url": file_url}


@responses.activate
def test_upload_dataset_from_url_with_forecast(dataset_with_forecast_point_server_data):
    prep_successful_upload_response(
        "urlUploads", json.dumps(dataset_with_forecast_point_server_data)
    )
    project = Project("p-id")
    file_url = "https://my_files.com/my_dataset.csv"
    forecast_point = parse_time(dataset_with_forecast_point_server_data["forecastPoint"])
    new_dataset = project.upload_dataset(file_url, forecast_point=forecast_point)

    assert new_dataset.forecast_point == forecast_point
    request = request_body_to_json(responses.calls[0].request)
    assert parse_time(request["forecastPoint"]) == forecast_point


@responses.activate
def test_upload_dataset_from_url_with_forecast_millis(
    dataset_with_millis_forecast_point_server_data,
):
    prep_successful_upload_response(
        "urlUploads", json.dumps(dataset_with_millis_forecast_point_server_data)
    )
    project = Project("p-id")
    file_url = "https://my_files.com/my_dataset.csv"
    forecast_point = parse_time(dataset_with_millis_forecast_point_server_data["forecastPoint"])
    new_dataset = project.upload_dataset(file_url, forecast_point=forecast_point)

    assert new_dataset.forecast_point == forecast_point
    assert new_dataset.forecast_point.microsecond == 800000
    request = request_body_to_json(responses.calls[0].request)
    assert parse_time(request["forecastPoint"]) == forecast_point


@responses.activate
def test_upload_dataset_from_url_with_config_id(dataset_with_config_id):
    prep_successful_upload_response("urlUploads", json.dumps(dataset_with_config_id))
    project = Project("p-id")
    file_url = "https://my_files.com/my_dataset.csv"
    config_id = dataset_with_config_id["secondaryDatasetsConfigId"]
    new_dataset = project.upload_dataset(file_url, secondary_datasets_config_id=config_id)

    assert new_dataset.secondary_datasets_config_id == config_id
    request = request_body_to_json(responses.calls[0].request)
    assert request["secondaryDatasetsConfigId"] == config_id


@responses.activate
def test_upload_dataset_from_file(dataset_json):
    prep_successful_upload_response("fileUploads", dataset_json)
    filepath = fixture_file_path("onehundredrows.csv")
    project = Project("p-id")
    new_dataset = project.upload_dataset(filepath)
    assert_matches_expected_dataset(new_dataset)
    assert "file" in responses.calls[0].request.body.fields


@responses.activate
def test_upload_dataset_with_dataset_filename(dataset_json):
    prep_successful_upload_response("fileUploads", dataset_json)
    project = Project("p-id")
    dataset_filename = "my_test_dataset"
    content = b"a,b,c\n" + b"1,2,3\n" * 100
    # Make a call to the API and provide a dataset_filename, rather than using default value.
    project.upload_dataset(content, dataset_filename=dataset_filename)
    fields = responses.calls[0].request.body.fields
    # Test that the provided dataset_filename was used in the request to api
    assert "file" in fields
    assert fields["file"][0] == dataset_filename


@responses.activate
def test_upload_dataset_from_file_with_config_id(dataset_with_config_id):
    prep_successful_upload_response("fileUploads", json.dumps(dataset_with_config_id))
    filepath = fixture_file_path("onehundredrows.csv")
    project = Project("p-id")
    config_id = dataset_with_config_id["secondaryDatasetsConfigId"]
    new_dataset = project.upload_dataset(filepath, secondary_datasets_config_id=config_id)

    assert new_dataset.secondary_datasets_config_id == config_id
    assert "secondaryDatasetsConfigId" in responses.calls[0].request.body.fields


@responses.activate
def test_upload_dataset_from_file_without_forecast_point(
    dataset_uploaded_without_forecast_point_server_data,
):
    prep_successful_upload_response(
        "fileUploads", json.dumps(dataset_uploaded_without_forecast_point_server_data)
    )

    forecast_point = parse_time(
        dataset_uploaded_without_forecast_point_server_data["forecastPoint"]
    )
    forecast_point_range = [
        parse_time(date)
        for date in dataset_uploaded_without_forecast_point_server_data["forecastPointRange"]
    ]
    data_start_date = parse_time(
        dataset_uploaded_without_forecast_point_server_data["dataStartDate"]
    )
    data_end_date = parse_time(dataset_uploaded_without_forecast_point_server_data["dataEndDate"])
    max_forecast_date = parse_time(
        dataset_uploaded_without_forecast_point_server_data["maxForecastDate"]
    )

    filepath = fixture_file_path("onehundredrows.csv")
    project = Project("p-id")
    new_dataset = project.upload_dataset(filepath)

    assert new_dataset.forecast_point == forecast_point
    assert new_dataset.forecast_point_range == forecast_point_range
    assert new_dataset.data_start_date == data_start_date
    assert new_dataset.data_end_date == data_end_date
    assert new_dataset.max_forecast_date == max_forecast_date

    assert len(responses.calls[0].request.body.fields.keys()) == 1


@responses.activate
def test_upload_dataset_from_file_with_forecast(dataset_with_forecast_point_server_data):
    prep_successful_upload_response(
        "fileUploads", json.dumps(dataset_with_forecast_point_server_data)
    )
    filepath = fixture_file_path("onehundredrows.csv")
    project = Project("p-id")
    forecast_point = parse_time(dataset_with_forecast_point_server_data["forecastPoint"])
    new_dataset = project.upload_dataset(filepath, forecast_point=forecast_point)

    assert new_dataset.forecast_point == forecast_point
    assert "forecastPoint" in responses.calls[0].request.body.fields


@responses.activate
def test_upload_dataset_from_file_with_forecast_millis(
    dataset_with_millis_forecast_point_server_data,
):
    prep_successful_upload_response(
        "fileUploads", json.dumps(dataset_with_millis_forecast_point_server_data)
    )
    filepath = fixture_file_path("onehundredrows.csv")
    project = Project("p-id")
    forecast_point = parse_time(dataset_with_millis_forecast_point_server_data["forecastPoint"])
    new_dataset = project.upload_dataset(filepath, forecast_point=forecast_point)

    assert new_dataset.forecast_point == forecast_point
    assert "forecastPoint" in responses.calls[0].request.body.fields


@responses.activate
def test_upload_dataset_from_url_with_start_end(dataset_with_start_end_server_data):
    prep_successful_upload_response("urlUploads", json.dumps(dataset_with_start_end_server_data))
    project = Project("p-id")
    file_url = "https://my_files.com/my_dataset.csv"
    predictions_start_date = parse_time(dataset_with_start_end_server_data["predictionsStartDate"])
    predictions_end_date = parse_time(dataset_with_start_end_server_data["predictionsEndDate"])

    new_dataset = project.upload_dataset(
        file_url,
        predictions_start_date=predictions_start_date,
        predictions_end_date=predictions_end_date,
        actual_value_column="actuals",
    )
    assert new_dataset.predictions_start_date == predictions_start_date
    assert new_dataset.predictions_end_date == predictions_end_date
    assert new_dataset.actual_value_column == "actuals"
    assert new_dataset.detected_actual_value_columns == [{"missing_count": 25, "name": "actuals"}]
    request = request_body_to_json(responses.calls[0].request)
    assert parse_time(request["predictionsStartDate"]) == predictions_start_date
    assert parse_time(request["predictionsEndDate"]) == predictions_end_date


@responses.activate
def test_upload_dataset_from_url_with_start_end_millis(dataset_with_millis_start_end_server_data):
    prep_successful_upload_response(
        "urlUploads", json.dumps(dataset_with_millis_start_end_server_data)
    )
    project = Project("p-id")
    file_url = "https://my_files.com/my_dataset.csv"
    predictions_start_date = parse_time(
        dataset_with_millis_start_end_server_data["predictionsStartDate"]
    )
    predictions_end_date = parse_time(
        dataset_with_millis_start_end_server_data["predictionsEndDate"]
    )

    new_dataset = project.upload_dataset(
        file_url,
        predictions_start_date=predictions_start_date,
        predictions_end_date=predictions_end_date,
    )
    assert new_dataset.predictions_start_date == predictions_start_date
    assert new_dataset.predictions_end_date == predictions_end_date
    request = request_body_to_json(responses.calls[0].request)
    assert parse_time(request["predictionsStartDate"]) == predictions_start_date
    assert parse_time(request["predictionsEndDate"]) == predictions_end_date


@responses.activate
def test_upload_dataset_from_url_fails_when_one_date_provided(dataset_with_start_end_server_data):
    project = Project("p-id")
    file_url = "https://my_files.com/my_dataset.csv"
    predictions_start_date = parse_time(dataset_with_start_end_server_data["predictionsStartDate"])
    msg = "Both prediction_start_date and prediction_end_date must be provided at the same time"
    with pytest.raises(ValueError) as e:
        project.upload_dataset(file_url, predictions_start_date=predictions_start_date)
    assert str(e.value) == msg


@responses.activate
def test_upload_dataset_from_url_fails_when_fp_and_dates_provided(
    dataset_with_start_end_server_data,
):
    project = Project("p-id")
    file_url = "https://my_files.com/my_dataset.csv"
    dataset_with_start_end_server_data["forecastPoint"] = "2017-01-01T15:00:00Z"
    forecast_point = parse_time(dataset_with_start_end_server_data["forecastPoint"])
    predictions_start_date = parse_time(dataset_with_start_end_server_data["predictionsStartDate"])
    predictions_end_date = parse_time(dataset_with_start_end_server_data["predictionsStartDate"])
    msg = (
        "forecast_point can not be provided together with predictions_start_date or "
        "predictions_end_date"
    )
    with pytest.raises(ValueError) as e:
        project.upload_dataset(
            file_url,
            forecast_point=forecast_point,
            predictions_start_date=predictions_start_date,
            predictions_end_date=predictions_end_date,
        )
    assert str(e.value) == msg


@responses.activate
def test_upload_dataset_from_file_with_start_end(dataset_with_start_end_server_data):
    prep_successful_upload_response("fileUploads", json.dumps(dataset_with_start_end_server_data))
    filepath = fixture_file_path("onehundredrows.csv")
    project = Project("p-id")
    predictions_start_date = parse_time(dataset_with_start_end_server_data["predictionsStartDate"])
    predictions_end_date = parse_time(dataset_with_start_end_server_data["predictionsEndDate"])

    new_dataset = project.upload_dataset(
        filepath,
        predictions_start_date=predictions_start_date,
        predictions_end_date=predictions_end_date,
    )
    assert new_dataset.predictions_start_date == predictions_start_date
    assert new_dataset.predictions_end_date == predictions_end_date
    assert "predictionsStartDate" in responses.calls[0].request.body.fields
    assert "predictionsEndDate" in responses.calls[0].request.body.fields


@responses.activate
def test_upload_dataset_from_file_with_start_end_millis(dataset_with_millis_start_end_server_data):
    prep_successful_upload_response(
        "fileUploads", json.dumps(dataset_with_millis_start_end_server_data)
    )
    filepath = fixture_file_path("onehundredrows.csv")
    project = Project("p-id")
    predictions_start_date = parse_time(
        dataset_with_millis_start_end_server_data["predictionsStartDate"]
    )
    predictions_end_date = parse_time(
        dataset_with_millis_start_end_server_data["predictionsEndDate"]
    )

    new_dataset = project.upload_dataset(
        filepath,
        predictions_start_date=predictions_start_date,
        predictions_end_date=predictions_end_date,
    )
    assert new_dataset.predictions_start_date == predictions_start_date
    assert new_dataset.predictions_end_date == predictions_end_date
    assert "predictionsStartDate" in responses.calls[0].request.body.fields
    assert "predictionsEndDate" in responses.calls[0].request.body.fields


@responses.activate
def test_upload_dataset_from_file_fails_when_one_date_provided(dataset_with_start_end_server_data):
    filepath = fixture_file_path("onehundredrows.csv")
    project = Project("p-id")
    predictions_start_date = parse_time(dataset_with_start_end_server_data["predictionsStartDate"])
    msg = "Both prediction_start_date and prediction_end_date must be provided at the same time"
    with pytest.raises(ValueError) as e:
        project.upload_dataset(filepath, predictions_start_date=predictions_start_date)
    assert str(e.value) == msg


@responses.activate
def test_upload_dataset_from_file_fails_when_fp_and_dates_provided(
    dataset_with_start_end_server_data,
):
    filepath = fixture_file_path("onehundredrows.csv")
    project = Project("p-id")
    dataset_with_start_end_server_data["forecastPoint"] = "2017-01-01T15:00:00Z"
    forecast_point = parse_time(dataset_with_start_end_server_data["forecastPoint"])
    predictions_end_date = parse_time(dataset_with_start_end_server_data["predictionsStartDate"])
    msg = (
        "forecast_point can not be provided together with predictions_start_date or "
        "predictions_end_date"
    )
    with pytest.raises(ValueError) as e:
        project.upload_dataset(
            filepath, forecast_point=forecast_point, predictions_end_date=predictions_end_date
        )
    assert str(e.value) == msg


@responses.activate
def test_upload_dataset_from_data_source(
    data_sources_endpoint,
    data_sources_list_server_resp,
    airlines10mb_data_source_server_resp,
    dataset_json,
):
    data_source_id = airlines10mb_data_source_server_resp["id"]
    username = "my_username"
    password = "my_password"

    prep_successful_upload_response("dataSourceUploads", dataset_json)
    responses.add(responses.GET, data_sources_endpoint, json=data_sources_list_server_resp)
    project = Project("p-id")

    project.upload_dataset_from_data_source(data_source_id, username=username, password=password)

    assert responses.calls[0].request.method == "POST"
    request_body = request_body_to_json(responses.calls[0].request)
    assert request_body == {"dataSourceId": data_source_id, "user": username, "password": password}


@responses.activate
def test_upload_dataset_from_data_source_with_forecast(
    data_sources_endpoint,
    data_sources_list_server_resp,
    airlines10mb_data_source_server_resp,
    dataset_with_forecast_point_server_data,
):
    data_source_id = airlines10mb_data_source_server_resp["id"]
    forecast_point = parse_time(dataset_with_forecast_point_server_data["forecastPoint"])
    username = "my_username"
    password = "my_password"
    prep_successful_upload_response(
        "dataSourceUploads", json.dumps(dataset_with_forecast_point_server_data)
    )
    responses.add(responses.GET, data_sources_endpoint, json=data_sources_list_server_resp)
    project = Project("p-id")

    project.upload_dataset_from_data_source(
        data_source_id, username, password, forecast_point=forecast_point
    )

    assert responses.calls[0].request.method == "POST"
    request_body = request_body_to_json(responses.calls[0].request)
    assert request_body == {
        "dataSourceId": data_source_id,
        "forecastPoint": datetime_to_string(forecast_point),
        "user": username,
        "password": password,
    }


@responses.activate
def test_upload_dataset_from_data_source_with_forecast_millis(
    data_sources_endpoint,
    data_sources_list_server_resp,
    airlines10mb_data_source_server_resp,
    dataset_with_millis_forecast_point_server_data,
):
    data_source_id = airlines10mb_data_source_server_resp["id"]
    forecast_point = parse_time(dataset_with_millis_forecast_point_server_data["forecastPoint"])
    username = "my_username"
    password = "my_password"
    prep_successful_upload_response(
        "dataSourceUploads", json.dumps(dataset_with_millis_forecast_point_server_data)
    )
    responses.add(responses.GET, data_sources_endpoint, json=data_sources_list_server_resp)
    project = Project("p-id")

    project.upload_dataset_from_data_source(
        data_source_id, username, password, forecast_point=forecast_point
    )

    assert responses.calls[0].request.method == "POST"
    request_body = request_body_to_json(responses.calls[0].request)
    assert request_body == {
        "dataSourceId": data_source_id,
        "forecastPoint": datetime_to_string(forecast_point),
        "user": username,
        "password": password,
    }


@responses.activate
def test_upload_dataset_from_data_source_without_forecast_point(
    data_sources_endpoint,
    data_sources_list_server_resp,
    airlines10mb_data_source_server_resp,
    dataset_uploaded_without_forecast_point_server_data,
):
    data_source_id = airlines10mb_data_source_server_resp["id"]
    forecast_point = parse_time(
        dataset_uploaded_without_forecast_point_server_data["forecastPoint"]
    )
    forecast_point_range = [
        parse_time(date)
        for date in dataset_uploaded_without_forecast_point_server_data["forecastPointRange"]
    ]
    data_start_date = parse_time(
        dataset_uploaded_without_forecast_point_server_data["dataStartDate"]
    )
    data_end_date = parse_time(dataset_uploaded_without_forecast_point_server_data["dataEndDate"])
    max_forecast_date = parse_time(
        dataset_uploaded_without_forecast_point_server_data["maxForecastDate"]
    )

    username = "my_username"
    password = "my_password"

    prep_successful_upload_response(
        "dataSourceUploads", json.dumps(dataset_uploaded_without_forecast_point_server_data)
    )
    responses.add(responses.GET, data_sources_endpoint, json=data_sources_list_server_resp)

    project = Project("p-id")
    new_dataset = project.upload_dataset_from_data_source(data_source_id, username, password)

    assert responses.calls[0].request.method == "POST"
    request_body = request_body_to_json(responses.calls[0].request)
    assert request_body == {
        "dataSourceId": data_source_id,
        "user": username,
        "password": password,
    }

    assert new_dataset.forecast_point == forecast_point
    assert new_dataset.forecast_point_range == forecast_point_range
    assert new_dataset.data_start_date == data_start_date
    assert new_dataset.data_end_date == data_end_date
    assert new_dataset.max_forecast_date == max_forecast_date
    assert new_dataset.actual_value_column is None
    assert new_dataset.detected_actual_value_columns == [{"missing_count": 25, "name": "actuals"}]


@responses.activate
def test_delete_dataset():
    responses.add(
        responses.DELETE, "https://host_name.com/projects/p-id/predictionDatasets/ds-id/", body=""
    )
    dataset = PredictionDataset.from_server_data(dataset_response)
    dataset.delete()
    assert responses.calls[0].request.method == "DELETE"
    assert len(responses.calls) == 1


@responses.activate
def test_upload_dataset_from_url_with_relax_kia_check(dataset_with_relax_kia_check_server_data):
    prep_successful_upload_response(
        "urlUploads", json.dumps(dataset_with_relax_kia_check_server_data)
    )
    project = Project("p-id")
    file_url = "https://my_files.com/my_dataset.csv"
    relax_kia_check = parse_time(
        dataset_with_relax_kia_check_server_data["relaxKnownInAdvanceFeaturesCheck"]
    )
    new_dataset = project.upload_dataset(
        file_url, relax_known_in_advance_features_check=relax_kia_check
    )

    assert new_dataset.relax_known_in_advance_features_check == relax_kia_check
    request = request_body_to_json(responses.calls[0].request)
    assert request["relaxKnownInAdvanceFeaturesCheck"] == str(relax_kia_check)


@responses.activate
def test_upload_dataset_from_file_with_relax_kia_check(dataset_with_relax_kia_check_server_data):
    prep_successful_upload_response(
        "fileUploads", json.dumps(dataset_with_relax_kia_check_server_data)
    )
    filepath = fixture_file_path("onehundredrows.csv")
    project = Project("p-id")
    relax_kia_check = parse_time(
        dataset_with_relax_kia_check_server_data["relaxKnownInAdvanceFeaturesCheck"]
    )
    new_dataset = project.upload_dataset(
        filepath, relax_known_in_advance_features_check=relax_kia_check
    )

    assert new_dataset.relax_known_in_advance_features_check == relax_kia_check
    assert "relaxKnownInAdvanceFeaturesCheck" in responses.calls[0].request.body.fields


@responses.activate
def test_upload_dataset_from_data_source_with_relax_kia_check(
    data_sources_endpoint,
    data_sources_list_server_resp,
    airlines10mb_data_source_server_resp,
    dataset_with_relax_kia_check_server_data,
):
    data_source_id = airlines10mb_data_source_server_resp["id"]
    relax_kia_check = parse_time(
        dataset_with_relax_kia_check_server_data["relaxKnownInAdvanceFeaturesCheck"]
    )
    username = "my_username"
    password = "my_password"
    prep_successful_upload_response(
        "dataSourceUploads", json.dumps(dataset_with_relax_kia_check_server_data)
    )
    responses.add(responses.GET, data_sources_endpoint, json=data_sources_list_server_resp)
    project = Project("p-id")

    project.upload_dataset_from_data_source(
        data_source_id, username, password, relax_known_in_advance_features_check=relax_kia_check
    )

    assert responses.calls[0].request.method == "POST"
    request_body = request_body_to_json(responses.calls[0].request)
    assert request_body == {
        "dataSourceId": data_source_id,
        "relaxKnownInAdvanceFeaturesCheck": relax_kia_check,
        "user": username,
        "password": password,
    }



# -*- coding: utf-8 -*-
from copy import deepcopy
import json

import pytest
import responses
import trafaret

from datarobot import errors
from datarobot.models.payoff_matrix import PayoffMatrix
from tests.utils import assert_urls_equal, request_body_to_json


@pytest.fixture
def payoff_matrix_id():
    return "5432deadbeefdeadbeef0000"


@pytest.fixture
def mocked_payoff_matrices(payoff_matrix_id, project_id):
    payoff_matrices_json = [
        {
            "id": payoff_matrix_id,
            "projectId": project_id,
            "name": "Test Matrix 1",
            "truePositiveValue": 1.0,
            "trueNegativeValue": 2.0,
            "falsePositiveValue": -3.0,
            "falseNegativeValue": -4.0,
        },
    ]
    return {
        "count": len(payoff_matrices_json),
        "next": None,
        "previous": None,
        "data": payoff_matrices_json,
    }


@pytest.fixture
def mocked_payoff_matrix(mocked_payoff_matrices):
    return mocked_payoff_matrices["data"][0]


@pytest.fixture
def make_payoff_matrices_url(unittest_endpoint, project_id):
    def _make_payoff_matrices_url(payoff_matrix_id=None):
        base_url = unittest_endpoint + "/projects/{}/payoffMatrices/".format(project_id)
        if payoff_matrix_id is not None:
            return "{}{}/".format(base_url, payoff_matrix_id)
        return base_url

    return _make_payoff_matrices_url


def mock_get_response(url, response, status=200, content_type="application/json"):
    if content_type == "application/json":
        body = json.dumps(response)
    else:
        body = response
    responses.add(responses.GET, url, status=status, content_type=content_type, body=body)


def assert_payoff_matrix(payoff_matrix, payoff_matrix_json):
    assert payoff_matrix.id == payoff_matrix_json["id"]
    assert payoff_matrix.project_id == payoff_matrix_json["projectId"]
    assert payoff_matrix.name == payoff_matrix_json["name"]
    assert payoff_matrix.true_positive_value == payoff_matrix_json["truePositiveValue"]
    assert payoff_matrix.true_negative_value == payoff_matrix_json["trueNegativeValue"]
    assert payoff_matrix.false_positive_value == payoff_matrix_json["falsePositiveValue"]
    assert payoff_matrix.false_negative_value == payoff_matrix_json["falseNegativeValue"]


def test_from_server_data(mocked_payoff_matrix):
    payoff_matrix = PayoffMatrix.from_server_data(mocked_payoff_matrix)
    assert_payoff_matrix(payoff_matrix, mocked_payoff_matrix)


@responses.activate
def test_list(project_id, mocked_payoff_matrices, make_payoff_matrices_url):
    url = make_payoff_matrices_url()
    mock_get_response(url, mocked_payoff_matrices)

    payoff_matrices = PayoffMatrix.list(project_id)
    assert payoff_matrices is not None
    assert len(mocked_payoff_matrices["data"]) == len(payoff_matrices)
    for payoff_matrix, payoff_matrix_json in zip(payoff_matrices, mocked_payoff_matrices["data"]):
        assert_payoff_matrix(payoff_matrix, payoff_matrix_json)
    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url


@responses.activate
def test_list_multiple_pages(project_id, mocked_payoff_matrices, make_payoff_matrices_url):
    url1 = make_payoff_matrices_url()
    url2 = make_payoff_matrices_url() + "2"

    mocked_payoff_matrices_2nd = deepcopy(mocked_payoff_matrices)
    mocked_payoff_matrices["next"] = url2

    mock_get_response(url1, mocked_payoff_matrices)
    mock_get_response(url2, mocked_payoff_matrices_2nd)

    payoff_matrices = PayoffMatrix.list(project_id)
    assert len(payoff_matrices) == len(mocked_payoff_matrices["data"]) + len(
        mocked_payoff_matrices_2nd["data"]
    )
    for payoff_matrix, payoff_matrix_json in zip(
        payoff_matrices, mocked_payoff_matrices["data"] + mocked_payoff_matrices_2nd["data"],
    ):
        assert_payoff_matrix(payoff_matrix, payoff_matrix_json)

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url.endswith(url1)
    assert responses.calls[1].request.method == "GET"
    assert responses.calls[1].request.url.endswith(url2)


@responses.activate
def test_list__no_id_in_api_response_causes_error(
    project_id, mocked_payoff_matrices, make_payoff_matrices_url
):
    mocked_payoff_matrices = deepcopy(mocked_payoff_matrices)
    payoff_matrix_no_id = {
        "id": None,
        "projectId": project_id,
        "modelId": "model-id",
        "datasetId": "dataset-id",
    }
    mocked_payoff_matrices["data"].append(payoff_matrix_no_id)

    url = make_payoff_matrices_url()
    mock_get_response(url, mocked_payoff_matrices)

    with pytest.raises(trafaret.DataError):
        PayoffMatrix.list(project_id)


@responses.activate
def test_create(one_model, project_url, payoff_matrix_id):
    payoff_matrices_url = project_url + "payoffMatrices/"

    submit_payoff_matrix = {
        "projectId": one_model.project_id,
        "id": payoff_matrix_id,
        "name": "some name",
        "truePositiveValue": 1,
        "trueNegativeValue": 2,
        "falsePositiveValue": -3,
        "falseNegativeValue": -4,
    }
    responses.add(
        responses.POST,
        payoff_matrices_url,
        body=json.dumps(submit_payoff_matrix),
        status=201,
        content_type="application/json",
    )

    payoff_matrix_created = PayoffMatrix.create(
        project_id=one_model.project_id, name=submit_payoff_matrix["name"],
    )

    request_url = responses.calls[0].request.url
    assert request_url == payoff_matrices_url

    request_json = request_body_to_json(responses.calls[0].request)
    assert request_json.get("truePositiveValue") == submit_payoff_matrix["truePositiveValue"]
    assert request_json.get("name") == submit_payoff_matrix["name"]
    assert request_json.get("name") == payoff_matrix_created.name


@responses.activate
def test_update(one_model, project_url, payoff_matrix_id):
    payoff_matrix_url = "{}payoffMatrices/{}/".format(project_url, payoff_matrix_id)

    submit_payoff_matrix = {
        "projectId": one_model.project_id,
        "id": payoff_matrix_id,
        "name": "some name",
        "truePositiveValue": 1,
        "trueNegativeValue": 2,
        "falsePositiveValue": -3,
        "falseNegativeValue": -4,
    }

    responses.add(
        responses.PUT,
        payoff_matrix_url,
        body=json.dumps(submit_payoff_matrix),
        status=200,
        content_type="application/json",
    )

    payoff_matrix_updated = PayoffMatrix.update(
        project_id=one_model.project_id,
        id=payoff_matrix_id,
        name=submit_payoff_matrix["name"],
        true_positive_value=1,
        true_negative_value=2,
        false_positive_value=-3,
        false_negative_value=-4,
    )

    request_json = request_body_to_json(responses.calls[0].request)
    assert request_json.get("truePositiveValue") == submit_payoff_matrix["truePositiveValue"]
    assert request_json.get("name") == submit_payoff_matrix["name"]
    assert request_json.get("name") == payoff_matrix_updated.name


@responses.activate
def test_delete(one_model, project_url, payoff_matrix_id):
    payoff_matrix_url = "{}payoffMatrices/{}/".format(project_url, payoff_matrix_id)
    print(payoff_matrix_url)
    responses.add(
        responses.DELETE, payoff_matrix_url, body="", status=204,
    )

    PayoffMatrix.delete(
        project_id=one_model.project_id, id=payoff_matrix_id,
    )

    request_url = responses.calls[0].request.url
    assert_urls_equal(request_url, payoff_matrix_url)


@responses.activate
def test_create__client_error(project_url, one_model):
    error_message = {"message": "some error message"}
    responses.add(
        responses.POST,
        project_url + "payoffMatrices/",
        body=json.dumps(error_message),
        status=422,
        content_type="application/json",
    )
    with pytest.raises(errors.ClientError):
        PayoffMatrix.create(project_id=one_model.project_id, name="some name")


@responses.activate
def test_create__server_error(project_url, one_model, dataset_id):
    payoff_matrices_url = project_url + "payoffMatrices/"
    submit_payoff_matrix = {
        "name": "some name",
        "truePositiveValue": 1,
        "trueNegativeValue": 2,
        "falsePositiveValue": -3,
        "falseNegativeValue": -4,
    }
    responses.add(
        responses.POST,
        payoff_matrices_url,
        body=json.dumps(submit_payoff_matrix),
        status=500,
        content_type="application/json",
    )
    with pytest.raises(errors.ServerError):
        PayoffMatrix.create(project_id=one_model.project_id, name="some name")



from __future__ import division

from datetime import datetime
import math
import random

import pandas as pd
import pytest
import responses

from datarobot import Deployment
from tests.utils import request_body_to_json


class TestSubmitActuals(object):
    @pytest.fixture()
    def response(self, unittest_endpoint, deployment_data):
        url = "{}/{}/{}/actuals/fromJSON/".format(
            unittest_endpoint, "deployments", deployment_data["id"]
        )
        status_url = "{}/STATUS_ID/".format(unittest_endpoint)
        accuracy_url = "{}/{}/{}/accuracy/".format(
            unittest_endpoint, "deployments", deployment_data["id"]
        )
        responses.add(responses.POST, url, status=202, headers={"Location": status_url})
        responses.add(responses.GET, status_url, status=303, headers={"Location": accuracy_url})

    @responses.activate
    @pytest.mark.usefixtures("deployment_get_response")
    @pytest.mark.parametrize("actuals_data", [None, [], "abc", pd.DataFrame()])
    def test_bad_data(self, deployment_data, actuals_data):
        """When submitting actuals with bad data, `ValueError` is raised."""
        with pytest.raises(ValueError):
            deployment = Deployment.get(deployment_data["id"])
            deployment.submit_actuals(actuals_data)

    @responses.activate
    @pytest.mark.usefixtures("deployment_get_response")
    @pytest.mark.parametrize("batch_size", [None, 3.7, 0, -3])
    def test_bad_batch_size(self, deployment_data, batch_size):
        """When submitting actuals with bad batch_size, `ValueError` is raised."""
        with pytest.raises(ValueError):
            deployment = Deployment.get(deployment_data["id"])
            deployment.submit_actuals(
                [{"association_id": "abc", "actual_value": "True"}], batch_size=batch_size,
            )

    @responses.activate
    @pytest.mark.usefixtures("deployment_get_response", "response")
    @pytest.mark.parametrize(
        "actuals_data, expected_payload",
        [
            (
                [{"association_id": "abc", "actual_value": "True"}],
                {"data": [{"associationId": "abc", "actualValue": "True"}]},
            ),
            (
                [{"association_id": "abc", "actual_value": "True"}],
                {"data": [{"associationId": "abc", "actualValue": "True"}]},
            ),
            (
                [{"association_id": "abc", "actual_value": "True", "was_acted_on": True}],
                {"data": [{"associationId": "abc", "actualValue": "True", "wasActedOn": True}]},
            ),
            (
                [
                    {
                        "association_id": "abc",
                        "actual_value": "True",
                        "timestamp": datetime(2019, 8, 1, 12, 30),
                    }
                ],
                {
                    "data": [
                        {
                            "associationId": "abc",
                            "actualValue": "True",
                            "timestamp": "2019-08-01T12:30:00+00:00",
                        }
                    ]
                },
            ),
        ],
    )
    def test_submit(self, deployment_data, actuals_data, expected_payload):
        deployment = Deployment.get(deployment_data["id"])
        deployment.submit_actuals(actuals_data)
        actual_payload = request_body_to_json(responses.calls[1].request)
        assert actual_payload == expected_payload

    @responses.activate
    @pytest.mark.usefixtures("deployment_get_response", "response")
    def test_submit_with_dataframe(self, deployment_data):
        actuals_data = pd.DataFrame(
            [
                ["a", 1, None, None],
                ["b", 2, True, None],
                ["c", 3, None, datetime(2019, 8, 1, 12, 30)],
                ["d", 4, False, datetime(2019, 8, 1, 12, 30)],
            ],
            columns=["association_id", "actual_value", "was_acted_on", "timestamp"],
        )
        expected_payload = {
            "data": [
                {"associationId": "a", "actualValue": 1},
                {"associationId": "b", "actualValue": 2, "wasActedOn": True},
                {"associationId": "c", "actualValue": 3, "timestamp": "2019-08-01T12:30:00+00:00"},
                {
                    "associationId": "d",
                    "actualValue": 4,
                    "wasActedOn": False,
                    "timestamp": "2019-08-01T12:30:00+00:00",
                },
            ]
        }

        deployment = Deployment.get(deployment_data["id"])
        deployment.submit_actuals(actuals_data)
        actual_payload = request_body_to_json(responses.calls[1].request)
        assert actual_payload == expected_payload

    @responses.activate
    @pytest.mark.usefixtures("deployment_get_response", "response")
    def test_batched_submit(self, deployment_data):
        """Test submit actuals in batches"""

        total_actuals_count = 536
        batch_size = 100

        # generate some test data
        random.seed(1005)
        actuals = [
            {
                "association_id": "abc",
                "actual_value": random.randrange(0, 100),
                "was_acted_on": bool(random.getrandbits(1)),
            }
            for _ in range(total_actuals_count)
        ]
        deployment = Deployment.get(deployment_data["id"])
        deployment.submit_actuals(actuals, batch_size=batch_size)

        # submit actuals
        post_request_count = 0
        payload_data = []
        for call in responses.calls:
            if call.request.method == "POST":
                post_request_count += 1
                payload = request_body_to_json(call.request)
                payload_data += payload["data"]

        # test if API calls are made correctly
        assert post_request_count == math.ceil(total_actuals_count / batch_size)
        assert len(actuals) == len(payload_data)
        for actual, payload in zip(actuals, payload_data):
            assert actual["association_id"] == payload["associationId"]
            assert actual["actual_value"] == payload["actualValue"]
            assert actual["was_acted_on"] == payload["wasActedOn"]



import json

import pytest
import responses

from datarobot import Project
from datarobot.enums import JOB_TYPE
from datarobot.errors import ClientError, InputNotUnderstoodError, JobAlreadyRequested
from datarobot.models.rating_table import RatingTable


@pytest.fixture
def rating_table_url(project_collection_url, rating_table_backend_generated_json):
    d = json.loads(rating_table_backend_generated_json)
    return "{}{}/ratingTables/{}/".format(project_collection_url, d["projectId"], d["id"])


@pytest.fixture
def downloaded_rating_table_contents():
    return b"""Matrix of word-grams occurrences tokenizer: scikit-learn based tokenizer
Matrix of word-grams occurrences binary: True
Matrix of word-grams occurrences sublinear_tf: False
Matrix of word-grams occurrences use_idf: False
Matrix of word-grams occurrences norm: L2
Intercept: -0.35839674349
Model precision: single
Loss distribution: Binomial Deviance
Link function: logit
Pairwise interactions found:
( number_emergency & number_inpatient ),0.0065962098320561146
( glyburide & number_emergency ),0.0016387530180316434

Feature Name,Feature Strength,Type,Transform1,Value1,Transform2,Value2,Weight,Coefficient
A1Cresult,0.0012348212995812102,CAT,One-hot,'>7',,,58.0,0.0082157774356443683
A1Cresult,0.0012348212995812102,CAT,One-hot,'>8',,,129.0,-0.012579536153936071
A1Cresult,0.0012348212995812102,CAT,One-hot,'Norm',,,59.0,0.032774696491885628
A1Cresult,0.0012348212995812102,CAT,One-hot,Missing value,,,1354.0,-0.00029931592903424328"""


@pytest.fixture
def downloaded_rating_table_file(temporary_file, downloaded_rating_table_contents):
    with open(temporary_file, "wb") as out_f:
        out_f.write(downloaded_rating_table_contents)
    return temporary_file


@pytest.fixture
def completed_rating_table_validation_job(job_id, project_id, job_url):
    return {
        "status": "COMPLETED",
        "url": job_url,
        "id": job_id,
        "jobType": JOB_TYPE.RATING_TABLE_VALIDATION,
        "isBlocked": False,
        "projectId": project_id,
    }


@responses.activate
def test_get_rating_table_backend_generated(
    rating_table_backend_generated_json, rating_table_backend_generated
):
    backend_generated_id = rating_table_backend_generated.id
    url = "https://host_name.com/projects/p-id/ratingTables/{}/".format(backend_generated_id)

    responses.add(
        responses.GET,
        url,
        body=rating_table_backend_generated_json,
        status=200,
        content_type="application/json",
    )

    table = RatingTable.get("p-id", backend_generated_id)
    assert isinstance(table, RatingTable)
    assert table.id == backend_generated_id
    assert not table.validation_job_id
    assert table.model_id


@responses.activate
def test_get_rating_table_not_modeled(
    rating_table_uploaded_but_not_modeled_json, rating_table_uploaded_but_not_modeled
):
    not_modeled_id = rating_table_uploaded_but_not_modeled.id
    url = "https://host_name.com/projects/p-id/ratingTables/{}/".format(not_modeled_id)

    responses.add(
        responses.GET,
        url,
        body=rating_table_uploaded_but_not_modeled_json,
        status=200,
        content_type="application/json",
    )
    table = RatingTable.get("p-id", not_modeled_id)
    assert isinstance(table, RatingTable)
    assert table.id == not_modeled_id
    assert table.validation_job_id
    assert not table.model_id


@responses.activate
def test_get_rating_table_modeled(
    rating_table_uploaded_and_modeled_json, rating_table_uploaded_and_modeled
):
    modeled_id = rating_table_uploaded_and_modeled.id
    url = "https://host_name.com/projects/p-id/ratingTables/{}/".format(modeled_id)

    responses.add(
        responses.GET,
        url,
        body=rating_table_uploaded_and_modeled_json,
        status=200,
        content_type="application/json",
    )
    table = RatingTable.get("p-id", modeled_id)
    assert isinstance(table, RatingTable)
    assert table.id == modeled_id
    assert table.validation_job_id
    assert table.model_id


@responses.activate
@pytest.mark.usefixtures("known_warning")
def test_get_invalid_rating_table_is_fine_but_warns(
    invalid_rating_table_json, invalid_rating_table
):
    modeled_id = invalid_rating_table.id
    url = "https://host_name.com/projects/p-id/ratingTables/{}/".format(modeled_id)

    responses.add(
        responses.GET,
        url,
        body=invalid_rating_table_json,
        status=200,
        content_type="application/json",
    )
    table = RatingTable.get("p-id", modeled_id)
    assert isinstance(table, RatingTable)
    assert table.id == modeled_id
    assert table.validation_job_id
    assert table.model_id


@responses.activate
@pytest.mark.usefixtures("known_warning")
def test_get_invalid_rating_table_from_job_warns(
    project_url,
    rating_table_url,
    invalid_rating_table_json,
    downloaded_rating_table_file,
    invalid_rating_table,
    completed_rating_table_validation_job,
    job_url,
):
    server_data = json.loads(invalid_rating_table_json)
    create_url = "{}ratingTables/".format(project_url)
    responses.add(
        responses.POST,
        create_url,
        body="",
        status=202,
        content_type="application/json",
        adding_headers={"Location": job_url},
    )
    responses.add(
        responses.GET,
        job_url,
        json=completed_rating_table_validation_job,
        status=303,
        adding_headers={"Location": rating_table_url},
    )
    responses.add(
        responses.GET,
        rating_table_url,
        body=invalid_rating_table_json,
        status=200,
        content_type="application/json",
    )
    job = RatingTable.create(
        server_data["projectId"], invalid_rating_table.model_id, downloaded_rating_table_file
    )
    table = job.get_result_when_complete()
    assert isinstance(table, RatingTable)
    assert table.validation_job_id
    assert table.model_id


@responses.activate
def test_list_with_invalid_rating_table_does_not_warn(project_url, invalid_rating_table_json):
    server_data = json.loads(invalid_rating_table_json)
    list_url = "{}ratingTables/".format(project_url)
    responses.add(
        responses.GET,
        list_url,
        body=json.dumps({"data": [server_data]}),
        status=200,
        content_type="application/json",
    )
    rating_tables = Project(server_data["projectId"]).get_rating_tables()
    table = rating_tables[0]
    assert isinstance(table, RatingTable)
    assert table.validation_job_id
    assert table.model_id


@responses.activate
def test_get_rating_table_not_found(rating_table_uploaded_and_modeled):
    modeled_id = rating_table_uploaded_and_modeled.id
    url = "https://host_name.com/projects/p-id/ratingTables/{}/".format(modeled_id)

    responses.add(
        responses.GET,
        url,
        body=json.dumps({"message": "This resource does not exist."}),
        status=404,
        content_type="application/json",
    )
    with pytest.raises(ClientError):
        RatingTable.get("p-id", modeled_id)


@responses.activate
def test_download_model_export(
    temporary_file, downloaded_rating_table_contents, rating_table_uploaded_and_modeled
):
    pid = rating_table_uploaded_and_modeled.project_id
    id = rating_table_uploaded_and_modeled.id
    url = "https://host_name.com/projects/{}/ratingTables/{}/file".format(pid, id)

    responses.add(responses.GET, url, body=downloaded_rating_table_contents)

    rating_table_uploaded_and_modeled.download(temporary_file)

    with open(temporary_file, "rb") as in_f:
        saved_code = in_f.read()
    assert saved_code == downloaded_rating_table_contents


@responses.activate
def test_create(
    project_url,
    base_job_server_data,
    rating_table_url,
    downloaded_rating_table_file,
    rating_table_backend_generated,
):
    url = "{}ratingTables/".format(project_url)
    job_url = "{}jobs/{}/".format(project_url, base_job_server_data["id"])
    responses.add(
        responses.POST,
        url,
        body="",
        status=202,
        content_type="application/json",
        adding_headers={"Location": job_url},
    )
    responses.add(
        responses.GET,
        job_url,
        json=base_job_server_data,
        status=303,
        adding_headers={"Location": rating_table_url},
    )

    job = RatingTable.create(
        base_job_server_data["projectId"],
        rating_table_backend_generated.model_id,
        downloaded_rating_table_file,
    )
    assert str(job.id) == base_job_server_data["id"]


@responses.activate
def test_create_from_file(
    project_url,
    base_job_server_data,
    rating_table_url,
    rating_table_backend_generated,
    downloaded_rating_table_file,
):
    url = "{}ratingTables/".format(project_url)
    job_url = "{}jobs/{}/".format(project_url, base_job_server_data["id"])
    responses.add(
        responses.POST,
        url,
        body="",
        status=202,
        content_type="application/json",
        adding_headers={"Location": job_url},
    )
    responses.add(
        responses.GET,
        job_url,
        json=base_job_server_data,
        status=303,
        adding_headers={"Location": rating_table_url},
    )

    job = RatingTable.create(
        base_job_server_data["projectId"],
        rating_table_backend_generated.model_id,
        downloaded_rating_table_file,
    )
    assert str(job.id) == base_job_server_data["id"]


@responses.activate
def test_create_invalid_sourcedata(base_job_server_data, rating_table_backend_generated):
    with pytest.raises(InputNotUnderstoodError):
        RatingTable.create(
            base_job_server_data["projectId"],
            rating_table_backend_generated.model_id,
            "Not real Rating Table content",
        )


@responses.activate
def test_create_bad_parent_model(project_url, base_job_server_data, downloaded_rating_table_file):
    url = "{}ratingTables/".format(project_url)
    responses.add(
        responses.POST,
        url,
        body=json.dumps({"message": "No existing rating table in the model"}),
        status=400,
        content_type="application/json",
    )

    with pytest.raises(ClientError):
        RatingTable.create(
            base_job_server_data["projectId"],
            "999999999999999999999999",
            downloaded_rating_table_file,
        )


@responses.activate
def test_create_model(
    rating_table_uploaded_but_not_modeled, base_job_server_data, rating_table_model_url, project_url
):
    url = "{}ratingTableModels/".format(project_url)
    job_url = "{}jobs/{}/".format(project_url, base_job_server_data["id"])
    responses.add(
        responses.POST,
        url,
        body="",
        status=202,
        content_type="application/json",
        adding_headers={"Location": job_url},
    )
    responses.add(
        responses.GET,
        job_url,
        json=base_job_server_data,
        status=303,
        adding_headers={"Location": rating_table_model_url},
    )

    job = rating_table_uploaded_but_not_modeled.create_model()
    assert str(job.id) == base_job_server_data["id"]


@responses.activate
def test_create_model_from_invalid_rating_table(
    project_url, base_job_server_data, rating_table_uploaded_but_not_modeled
):
    url = "{}ratingTableModels/".format(project_url)
    job_url = "{}jobs/{}/".format(project_url, base_job_server_data["id"])
    responses.add(
        responses.POST,
        url,
        body=json.dumps(
            {"message": "A prerequisite was not satisfied: Rating Table failed validation"}
        ),
        status=422,
        content_type="application/json",
        adding_headers={"Location": job_url},
    )

    with pytest.raises(ClientError):
        rating_table_uploaded_but_not_modeled.create_model()


@responses.activate
def test_create_model_from_used_rating_table(
    project_url, base_job_server_data, rating_table_uploaded_and_modeled
):
    url = "{}ratingTableModels/".format(project_url)
    job_url = "{}jobs/{}/".format(project_url, base_job_server_data["id"])
    responses.add(
        responses.POST,
        url,
        body=json.dumps(
            {"message": "Rating Table already has a model", "errorName": "JobAlreadyAdded"}
        ),
        status=422,
        content_type="application/json",
        adding_headers={"Location": job_url},
    )

    with pytest.raises(JobAlreadyRequested):
        rating_table_uploaded_and_modeled.create_model()


@responses.activate
def test_rename_rating_table(
    rating_table_uploaded_but_not_modeled, rating_table_renamed_json, renamed_rating_table
):
    project_id = renamed_rating_table.project_id
    table_id = renamed_rating_table.id
    url = "https://host_name.com/projects/{}/ratingTables/{}/".format(project_id, table_id)
    responses.add(
        responses.PATCH,
        url,
        body=rating_table_renamed_json,
        status=202,
        content_type="application/json",
    )
    table = rating_table_uploaded_but_not_modeled
    assert isinstance(table, RatingTable)
    assert table.rating_table_name != "renamed"
    table.rename("renamed")
    assert table.rating_table_name == "renamed"
